"use strict";
// Copyright (c) 2020 Brandon Lehmann
//
// Please see the included LICENSE file for more information.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MySQL = void 0;
const events_1 = require("events");
const Types_1 = require("./Types");
const mysql_1 = require("mysql");
const util_1 = require("util");
/**
 * MYSQL interface that implements IDatabase
 */
class MySQL extends events_1.EventEmitter {
    /**
     * Creates a new instance of the class
     * @param host
     * @param port
     * @param user
     * @param password
     * @param database
     * @param connectionLimit
     */
    constructor(host, port, user, password, database, connectionLimit = 10) {
        super();
        this.m_tableOptions = 'ENGINE=InnoDB PACK_KEYS=1 ROW_FORMAT=COMPRESSED';
        this.m_db = mysql_1.createPool({ host, port, user, password, database, connectionLimit });
        this.m_db.on('error', error => this.emit('error', error));
        this.m_db.on('acquire', connection => this.emit('acquire', connection));
        this.m_db.on('connection', connection => this.emit('connection', connection));
        this.m_db.on('enqueue', () => this.emit('enqueue'));
        this.m_db.on('release', connection => this.emit('release', connection));
        this.emit('ready');
    }
    get hashType() {
        return 'char(64)';
    }
    get blobType() {
        return 'longtext';
    }
    get uint32Type() {
        return 'int(10) unsigned';
    }
    get uint64Type() {
        return 'bigint(20) unsigned';
    }
    get tableOptions() {
        return this.m_tableOptions;
    }
    set tableOptions(value) {
        this.m_tableOptions = value;
    }
    get type() {
        return Types_1.Interfaces.DBType.MYSQL;
    }
    on(event, listener) {
        return super.on(event, listener);
    }
    /**
     * Creates the table with the given fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    createTable(name, fields, primaryKey, tableOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield Types_1.createTable(this, this.type, name, fields, primaryKey, tableOptions);
            }
            catch (error) {
                this.emit('error', error);
            }
        });
    }
    /**
     * Creates the SQL statements necessary to create a table with the given
     * fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    prepareCreateTable(name, fields, primaryKey, tableOptions) {
        return Types_1.prepareCreateTable(this.type, name, fields, primaryKey, tableOptions);
    }
    /**
     * Closes the database connection(s)
     */
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_db.end(err => {
                    if (err)
                        return reject(err);
                    return resolve();
                });
            });
        });
    }
    /**
     * Performs a query and returns the result
     * @param query
     * @param values
     * @param openClient permit specifying the pool client to use
     */
    query(query, values) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_db.query(query, values || [], (error, results) => {
                    if (error)
                        return reject(error);
                    const count = results.changedRows ||
                        results.affectedRows ||
                        results.insertId ||
                        results.length;
                    return resolve([count, results]);
                });
            });
        });
    }
    /**
     * Constructs and executes the given queries in a transaction
     * @param queries
     */
    transaction(queries) {
        return __awaiter(this, void 0, void 0, function* () {
            const connection = yield this.connection();
            try {
                yield beginTransaction(connection);
                for (const q of queries) {
                    yield query(connection, q.query, q.values);
                }
                yield commit(connection);
            }
            catch (e) {
                yield rollback(connection);
                throw e;
            }
            finally {
                connection.release();
            }
        });
    }
    /**
     * Prepares a query to perform a multi-insert statement which is
     * far faster than a whole bunch of individual insert statements
     * @param table the table to perform the insert on
     * @param columns the columns to include in the insert
     * @param values the value of the columns
     */
    prepareMultiInsert(table, columns, values) {
        const query = util_1.format('INSERT INTO %s (%s) VALUES %L', table, columns.join(','));
        if (values) {
            const escaped = mysql_1.escape(values);
            return query.replace('%L', escaped);
        }
        return query;
    }
    /**
     * Prepares a query to perform a multi-update statement which is
     * based upon a multi-insert statement that performs an UPSERT
     * and this is a lot faster than a whole bunch of individual
     * update statements
     * @param table the table to perform the insert on
     * @param primaryKey the primary key(s) of the column for update purposes
     * @param columns the columns to update in the table
     * @param values the value of the columns
     */
    prepareMultiUpdate(table, primaryKey, columns, values) {
        const query = this.prepareMultiInsert(table, primaryKey.concat(columns), values);
        const updates = [];
        for (const column of columns) {
            updates.push(util_1.format(' %s = VALUES(%s)', column, column));
        }
        return util_1.format('%s ON DUPLICATE KEY UPDATE %s', query, updates.join(','));
    }
    connection() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_db.getConnection((error, connection) => {
                    if (error)
                        return reject(error);
                    return resolve(connection);
                });
            });
        });
    }
}
exports.MySQL = MySQL;
/** @ignore */
function beginTransaction(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.beginTransaction(error => {
                if (error)
                    return reject(error);
                return resolve();
            });
        });
    });
}
/** @ignore */
function commit(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.commit(error => {
                if (error)
                    return reject(error);
                return resolve();
            });
        });
    });
}
/** @ignore */
function query(connection, query, values = []) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.query(query, values, (error) => {
                if (error)
                    return reject(error);
                return resolve();
            });
        });
    });
}
/** @ignore */
function rollback(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.rollback(error => {
                if (error)
                    return reject(error);
                return resolve();
            });
        });
    });
}
