"use strict";
// Copyright (c) 2020 Brandon Lehmann
//
// Please see the included LICENSE file for more information.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Postgres = void 0;
const events_1 = require("events");
const pg_1 = require("pg");
const pgformat = require("pg-format");
const util_1 = require("util");
const Types_1 = require("./Types");
/**
 * Postgres database abstraction interface
 */
class Postgres extends events_1.EventEmitter {
    /**
     * Creates a new instance of the class
     * @param host
     * @param port
     * @param username
     * @param password
     * @param database
     */
    constructor(host, port, username, password, database) {
        super();
        this.m_tableOptions = undefined;
        this.m_pool = new pg_1.Pool({
            user: username,
            password: password,
            host: host,
            port: port,
            database: database,
            ssl: {
                rejectUnauthorized: false
            }
        });
        this.m_pool.on('connect', client => this.emit('connect', client));
        this.m_pool.on('acquire', client => this.emit('acquire', client));
        this.m_pool.on('remove', client => this.emit('remove', client));
        this.m_pool.on('error', (error, client) => this.emit('error', error, client));
        this.emit('ready');
    }
    /**
     * Returns the column type for the DB for a hash (ie. char(64))
     */
    get hashType() {
        return 'char(64)';
    }
    /**
     * Returns the column type for the DB to store a hexadecimal blob of data
     */
    get blobType() {
        return 'text';
    }
    /**
     * Returns the column type for the DB to store a uint32_t value
     */
    get uint32Type() {
        return 'numeric(10)';
    }
    /**
     * Returns the column type for the DB to store a uint64_t value
     */
    get uint64Type() {
        return 'numeric(20)';
    }
    /**
     * Returns a string of additional options that need to be applied to tables
     * which may include ENGINE information, compression, etc depending on the DB system
     */
    get tableOptions() {
        return this.m_tableOptions;
    }
    set tableOptions(value) {
        this.m_tableOptions = value;
    }
    /**
     * Returns the database type
     */
    get type() {
        return Types_1.Interfaces.DBType.POSTGRES;
    }
    on(event, listener) {
        return super.on(event, listener);
    }
    /**
     * Closes the database connection(s)
     */
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.m_pool.end();
        });
    }
    /**
     * Creates the table with the given fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    createTable(name, fields, primaryKey, tableOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield Types_1.createTable(this, this.type, name, fields, primaryKey, tableOptions);
            }
            catch (error) {
                this.emit('error', error);
            }
        });
    }
    /**
     * Creates the SQL statements necessary to create a table with the given
     * fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    prepareCreateTable(name, fields, primaryKey, tableOptions) {
        return Types_1.prepareCreateTable(this.type, name, fields, primaryKey, tableOptions);
    }
    /**
     * Performs a query and returns the result
     * @param query
     * @param values
     * @param openClient permit specifying the pool client to use
     */
    query(query, values, openClient) {
        return __awaiter(this, void 0, void 0, function* () {
            const client = openClient || (yield this.m_pool.connect());
            query = transformQuery(query);
            const res = yield client.query(query, values);
            if (!openClient) {
                yield client.release();
            }
            return [res.rowCount, res.rows];
        });
    }
    /**
     * Constructs and executes the given queries in a transaction
     * @param queries
     */
    transaction(queries) {
        return __awaiter(this, void 0, void 0, function* () {
            const client = yield this.m_pool.connect();
            try {
                yield client.query('BEGIN');
                for (const q of queries) {
                    yield this.query(q.query, q.values, client);
                }
                yield client.query('COMMIT');
            }
            catch (e) {
                yield client.query('ROLLBACK');
                throw e;
            }
            finally {
                yield client.release();
            }
        });
    }
    /**
     * Prepares a query to perform a multi-insert statement which is
     * far faster than a whole bunch of individual insert statements
     * @param table the table to perform the insert on
     * @param columns the columns to include in the insert
     * @param values the value of the columns
     */
    prepareMultiInsert(table, columns, values) {
        const query = util_1.format('INSERT INTO %s (%s) VALUES %L', table, columns.join(','));
        return pgformat(query, values);
    }
    /**
     * Prepares a query to perform a multi-update statement which is
     * based upon a multi-insert statement that performs an UPSERT
     * and this is a lot faster than a whole bunch of individual
     * update statements
     * @param table the table to perform the insert on
     * @param primaryKey the primary key(s) of the column for update purposes
     * @param columns the columns to update in the table
     * @param values the value of the columns
     */
    prepareMultiUpdate(table, primaryKey, columns, values) {
        const query = this.prepareMultiInsert(table, primaryKey.concat(columns), values);
        const updates = [];
        for (const column of columns) {
            updates.push(util_1.format(' %s = excluded.%s', column, column));
        }
        return util_1.format('%s ON CONFLICT (%s) DO UPDATE SET %s', query, primaryKey.join(','), updates.join(','));
    }
}
exports.Postgres = Postgres;
/** @ignore */
function transformQuery(query) {
    let counter = 1;
    while (query.indexOf('?') !== -1) {
        query = query.replace(/\?/, `$${counter}`);
        counter++;
    }
    return query;
}
