/// <reference types="node" />
import { EventEmitter } from 'events';
import { IDatabase, Interfaces } from './Types';
import { Database } from 'sqlite3';
/**
 * Sqlite database abstraction interface
 */
export declare class SQLite extends EventEmitter implements IDatabase {
    private readonly m_db;
    private m_tableOptions;
    private m_timer;
    private m_transactionQueue;
    private m_runQueue;
    private m_allQueue;
    /**
     * Creates a new instance of the class
     * @param filepath
     * @param mode
     */
    constructor(filepath: string, mode?: number);
    /**
     * Returns the column type for the DB for a hash (ie. char(64))
     */
    get hashType(): string;
    /**
     * Returns the column type for the DB to store a hexadecimal blob of data
     */
    get blobType(): string;
    /**
     * Returns the column type for the DB to store a uint32_t value
     */
    get uint32Type(): string;
    /**
     * Returns the column type for the DB to store a uint64_t value
     */
    get uint64Type(): string;
    /**
     * Returns a string of additional options that need to be applied to tables
     * which may include ENGINE information, compression, etc depending on the DB system
     */
    get tableOptions(): string | undefined;
    set tableOptions(value: string | undefined);
    /**
     * Returns the database type
     */
    get type(): Interfaces.DBType;
    on(event: 'error', listener: (error: Error) => void): this;
    on(event: 'ready', listener: (filepath: string, mode: number) => void): this;
    /**
     * Creates the table with the given fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    createTable(name: string, fields: Interfaces.ITableColumn[], primaryKey: string[], tableOptions?: string): Promise<void>;
    /**
     * Creates the SQL statements necessary to create a table with the given
     * fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    prepareCreateTable(name: string, fields: Interfaces.ITableColumn[], primaryKey: string[], tableOptions?: string): {
        table: string;
        indexes: string[];
    };
    /**
     * Retrieves the current PRAGMA setting
     * @param option
     */
    getPragma(option: string): Promise<any>;
    /**
     * Sets the given PRAGMA setting
     * @param option
     * @param value
     */
    setPragma(option: string, value: boolean | number | string): Promise<void>;
    /**
     * Closes the database connection(s)
     */
    close(): Promise<void>;
    /**
     * Performs a query and returns the result
     * @param query
     * @param values
     */
    query(query: string, values?: any[]): Promise<Interfaces.IQueryResult>;
    /**
     * Returns the underlying database connection
     */
    connection(): Promise<Database>;
    /**
     * Constructs and executes the given queries in a transaction
     * @param queries
     */
    transaction(queries: Interfaces.IBulkQuery[]): Promise<void>;
    /**
     * Prepares a query to perform a multi-insert statement which is
     * far faster than a whole bunch of individual insert statements
     * @param table the table to perform the insert on
     * @param columns the columns to include in the insert
     * @param values the value of the columns
     */
    prepareMultiInsert(table: string, columns: string[], values?: Interfaces.IValueArray): string;
    /**
     * Prepares a query to perform a multi-update statement which is
     * based upon a multi-insert statement that performs an UPSERT
     * and this is a lot faster than a whole bunch of individual
     * update statements
     * @param table the table to perform the insert on
     * @param primaryKey the primary key(s) of the column for update purposes
     * @param columns the columns to update in the table
     * @param values the value of the columns
     */
    prepareMultiUpdate(table: string, primaryKey: string[], columns: string[], values?: Interfaces.IValueArray): string;
    private all;
    private run;
}
