"use strict";
// Copyright (c) 2020 Brandon Lehmann
//
// Please see the included LICENSE file for more information.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SQLite = void 0;
const events_1 = require("events");
const Types_1 = require("./Types");
const path_1 = require("path");
const sqlite3_1 = require("sqlite3");
const mysql_1 = require("mysql");
const node_metronome_1 = require("node-metronome");
const util_1 = require("util");
/** @ignore */
const pragmaFunctionCalls = [
    'quick_check',
    'integrity_check',
    'incremental_vacuum',
    'foreign_key_check',
    'foreign_key_list',
    'index_info',
    'index_list',
    'index_xinfo',
    'table_info',
    'table_xinfo',
    'optimize'
];
/**
 * Sqlite database abstraction interface
 */
class SQLite extends events_1.EventEmitter {
    /**
     * Creates a new instance of the class
     * @param filepath
     * @param mode
     */
    constructor(filepath, mode = sqlite3_1.OPEN_READWRITE | sqlite3_1.OPEN_CREATE) {
        super();
        this.m_tableOptions = undefined;
        this.m_transactionQueue = [];
        this.m_runQueue = [];
        this.m_allQueue = [];
        filepath = path_1.resolve(filepath);
        this.m_db = new sqlite3_1.Database(filepath, error => {
            if (error)
                return this.emit('error', error);
            this.emit('ready', filepath, mode);
        });
        this.on('ready', () => __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.setPragma('foreign_keys', true);
            }
            catch (e) {
                this.emit('error', e);
            }
        }));
        this.m_timer = new node_metronome_1.Metronome(250, true);
        this.m_timer.on('tick', () => __awaiter(this, void 0, void 0, function* () {
            this.m_timer.paused = true;
            while (this.m_transactionQueue.length > 0) {
                const queuedTransaction = this.m_transactionQueue.shift();
                if (!queuedTransaction) {
                    break;
                }
                const connection = yield this.connection();
                try {
                    yield beginTransaction(connection);
                    for (const q of queuedTransaction.queries) {
                        yield query(connection, q.query, q.values);
                    }
                    yield commit(connection);
                    queuedTransaction.callback(undefined, undefined);
                }
                catch (e) {
                    yield rollback(connection);
                    queuedTransaction.callback(e, undefined);
                }
            }
            while (this.m_allQueue.length > 0) {
                const queue = this.m_allQueue.shift();
                if (!queue) {
                    break;
                }
                try {
                    const result = yield this.all(queue.query, queue.values);
                    queue.callback(undefined, result);
                }
                catch (e) {
                    queue.callback(e, undefined);
                }
            }
            while (this.m_runQueue.length > 0) {
                const queue = this.m_runQueue.shift();
                if (!queue) {
                    break;
                }
                try {
                    const result = yield this.run(queue.query, queue.values);
                    queue.callback(undefined, result);
                }
                catch (e) {
                    queue.callback(e, undefined);
                }
            }
            this.m_timer.paused = false;
        }));
    }
    /**
     * Returns the column type for the DB for a hash (ie. char(64))
     */
    get hashType() {
        return 'varchar(64)';
    }
    /**
     * Returns the column type for the DB to store a hexadecimal blob of data
     */
    get blobType() {
        return 'text';
    }
    /**
     * Returns the column type for the DB to store a uint32_t value
     */
    get uint32Type() {
        return 'unsigned int';
    }
    /**
     * Returns the column type for the DB to store a uint64_t value
     */
    get uint64Type() {
        return 'unsigned bigint';
    }
    /**
     * Returns a string of additional options that need to be applied to tables
     * which may include ENGINE information, compression, etc depending on the DB system
     */
    get tableOptions() {
        return this.m_tableOptions;
    }
    set tableOptions(value) {
        this.m_tableOptions = value;
    }
    /**
     * Returns the database type
     */
    get type() {
        return Types_1.Interfaces.DBType.SQLITE;
    }
    on(event, listener) {
        return super.on(event, listener);
    }
    /**
     * Creates the table with the given fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    createTable(name, fields, primaryKey, tableOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield Types_1.createTable(this, this.type, name, fields, primaryKey, tableOptions);
            }
            catch (error) {
                this.emit('error', error);
            }
        });
    }
    /**
     * Creates the SQL statements necessary to create a table with the given
     * fields and options
     * @param name
     * @param fields
     * @param primaryKey
     * @param tableOptions
     */
    prepareCreateTable(name, fields, primaryKey, tableOptions) {
        return Types_1.prepareCreateTable(this.type, name, fields, primaryKey, tableOptions);
    }
    /**
     * Retrieves the current PRAGMA setting
     * @param option
     */
    getPragma(option) {
        return __awaiter(this, void 0, void 0, function* () {
            option = option.toLowerCase();
            const query = 'PRAGMA ' + option;
            const [count, rows] = yield this.query(query);
            if (count === 1) {
                return rows[0][option];
            }
            else {
                if (rows[0][option]) {
                    return rows.map(elem => elem[option]);
                }
            }
            return rows;
        });
    }
    /**
     * Sets the given PRAGMA setting
     * @param option
     * @param value
     */
    setPragma(option, value) {
        return __awaiter(this, void 0, void 0, function* () {
            option = option.toLowerCase();
            let query = 'PRAGMA ';
            if (pragmaFunctionCalls.indexOf(option) !== -1) {
                query += option + '(' + value + ')';
            }
            else {
                query += option + ' = ' + value;
            }
            yield this.query(query);
        });
    }
    /**
     * Closes the database connection(s)
     */
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            // eslint-disable-next-line no-async-promise-executor
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                const sleep = (timeout = 100) => __awaiter(this, void 0, void 0, function* () { return new Promise(resolve => setTimeout(resolve, timeout)); });
                /**
                 * Let's not close the database connection while we still have stuff
                 * waiting to get done
                 */
                while (this.m_transactionQueue.length !== 0 ||
                    this.m_runQueue.length !== 0 ||
                    this.m_allQueue.length !== 0) {
                    yield sleep();
                }
                this.m_db.close(error => {
                    if (error)
                        return reject(error);
                    return resolve();
                });
            }));
        });
    }
    /**
     * Performs a query and returns the result
     * @param query
     * @param values
     */
    query(query, values) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                if (query.toLowerCase().startsWith('select')) {
                    this.m_allQueue.push({
                        query,
                        values,
                        callback: (error, result) => {
                            if (error) {
                                return reject(error);
                            }
                            return resolve(result || [0, []]);
                        }
                    });
                }
                else {
                    this.m_runQueue.push({
                        query,
                        values,
                        callback: (error, result) => {
                            if (error) {
                                return reject(error);
                            }
                            return resolve(result || [0, []]);
                        }
                    });
                }
            });
        });
    }
    /**
     * Returns the underlying database connection
     */
    connection() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.m_db;
        });
    }
    /**
     * Constructs and executes the given queries in a transaction
     * @param queries
     */
    transaction(queries) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_transactionQueue.push({
                    queries: queries,
                    callback: (error) => {
                        if (error) {
                            return reject(error);
                        }
                        return resolve();
                    }
                });
            });
        });
    }
    /**
     * Prepares a query to perform a multi-insert statement which is
     * far faster than a whole bunch of individual insert statements
     * @param table the table to perform the insert on
     * @param columns the columns to include in the insert
     * @param values the value of the columns
     */
    prepareMultiInsert(table, columns, values) {
        const query = util_1.format('INSERT INTO %s (%s) VALUES %L', table, columns.join(','));
        if (values) {
            const escaped = mysql_1.escape(values);
            return query.replace('%L', escaped);
        }
        return query;
    }
    /**
     * Prepares a query to perform a multi-update statement which is
     * based upon a multi-insert statement that performs an UPSERT
     * and this is a lot faster than a whole bunch of individual
     * update statements
     * @param table the table to perform the insert on
     * @param primaryKey the primary key(s) of the column for update purposes
     * @param columns the columns to update in the table
     * @param values the value of the columns
     */
    prepareMultiUpdate(table, primaryKey, columns, values) {
        const query = this.prepareMultiInsert(table, primaryKey.concat(columns), values);
        const updates = [];
        for (const column of columns) {
            updates.push(util_1.format(' %s = excluded.%s', column, column));
        }
        return util_1.format('%s ON CONFLICT (%s) DO UPDATE SET %s', query, primaryKey.join(','), updates.join(','));
    }
    all(query, values) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_db.all(query, values, (error, rows) => {
                    if (error) {
                        return reject(error);
                    }
                    return resolve([rows.length, rows]);
                });
            });
        });
    }
    run(query, values) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.m_db.run(query, values, function (error) {
                    if (error) {
                        return reject(error);
                    }
                    const count = (query.toLowerCase().startsWith('insert'))
                        ? this.lastID
                        : (query.toLowerCase().startsWith('update') ||
                            query.toLowerCase().startsWith('delete'))
                            ? this.changes
                            : 0;
                    return resolve([count, []]);
                });
            });
        });
    }
}
exports.SQLite = SQLite;
/** @ignore */
function beginTransaction(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.run('BEGIN TRANSACTION', err => {
                if (err)
                    return reject(err);
                return resolve();
            });
        });
    });
}
/** @ignore */
function commit(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.run('COMMIT', err => {
                if (err)
                    return reject(err);
                return resolve();
            });
        });
    });
}
/** @ignore */
function query(connection, query, values = []) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.run(query, values, err => {
                if (err)
                    return reject(err);
                return resolve();
            });
        });
    });
}
/** @ignore */
function rollback(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            connection.run('ROLLBACK', err => {
                if (err)
                    return reject(err);
                return resolve();
            });
        });
    });
}
